using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.Serialization.Formatters.Binary;
using UnityEngine;
using XNode;

namespace PlugPeaNodes
{
    [CreateNodeMenu("")]
    public class Curve : Node
    {

        public int type = 0;     //"0 = polyline, 1 = nurbs, 2= circle"
        public Vector3[] ControlPoints;
        public bool closed = false;
        public float radius;
        public Plane center;

    }


    //_____________________________________________________________

    public class Curves
    {
        public Curve DeepCopy()
        {
            using (MemoryStream memoryStream = new MemoryStream())
            {
                BinaryFormatter binaryFormatter = new BinaryFormatter();
                binaryFormatter.Serialize(memoryStream, this);
                memoryStream.Position = 0;
                return (Curve)binaryFormatter.Deserialize(memoryStream);
            }
        }
        public static class Transformation
        {
            public static Curve Move(Curve crv, Vector3 moveVector)
            {
                for(int i = 0; i < crv.ControlPoints.Length; i++)
                {
                    crv.ControlPoints[i] += moveVector;
                }
                if (crv.type == 2)
                {
                    crv.center.basePoint += moveVector;
                }
                return crv;
            }
          

        }
        public class CurveInterpolation
        {
            public static Vector3 GetPointOnCurve(Curve curve, int interpolationType, float t)
            {
                Vector3[] controlPoints = curve.ControlPoints;
                if (interpolationType == 0)
                {
                    // polyline interpolation
                    int index = Mathf.Min(Mathf.FloorToInt(t * controlPoints.Length), controlPoints.Length - 1);
                    float localT = t * controlPoints.Length - index;

                    return Vector3.Lerp(controlPoints[index], controlPoints[Mathf.Min(index + 1, controlPoints.Length - 1)], localT);
                }
                else if (interpolationType == 1)
                {
                    // bezier interpolation
                    return BezierInterpolation(controlPoints, t);
                }
                else if (interpolationType == 2)
                {
                    // circular interpolation
                    //    Debug.Log(curve.radius + " , " + curve.center.basePoint + " , " + t);
                    return PointOnCircle(curve.center, curve.radius, t);
                    //return new Vector3(0,0,0);
                }
                else
                {
                    throw new System.ArgumentException("Invalid interpolation type");
                }
            }

            public static Vector3 GetTangentOnCurve(Curve curve, int interpolationType, float t)
            {
                Vector3 tangent;
                Vector3[] controlPoints = curve.ControlPoints;
                if (interpolationType == 0)
                {
                    // polyline tangent
                    int index = Mathf.Min(Mathf.FloorToInt(t * controlPoints.Length), controlPoints.Length - 1);
                    tangent = controlPoints[Mathf.Min(index + 1, controlPoints.Length - 1)] - controlPoints[index];
                }
                else if (interpolationType == 1)
                {
                    // bezier tangent
                    Vector3 left = BezierInterpolation(controlPoints, Mathf.Max(0, t - 0.01f));
                    Vector3 right = BezierInterpolation(controlPoints, Mathf.Min(1, t + 0.01f));

                    tangent = (right - left).normalized;
                }
                else if (interpolationType == 2)
                {
                    // circular tangent
                    Vector3 currentPoint = PointOnCircle(curve.center, curve.radius, t );
                    Vector3 nextPoint = PointOnCircle(curve.center,  curve.radius, t + 0.01f);

                    tangent = (nextPoint - currentPoint).normalized;
                }
                else
                {
                    throw new System.ArgumentException("Invalid interpolation type");
                }

                return tangent;
            }

            private static Vector3 BezierInterpolation(Vector3[] controlPoints, float t)
            {
                Vector3[] tempPoints = new Vector3[controlPoints.Length];
                System.Array.Copy(controlPoints, tempPoints, controlPoints.Length);

                for (int i = controlPoints.Length - 1; i > 0; i--)
                {
                    for (int j = 0; j < i; j++)
                    {
                        tempPoints[j] = Vector3.Lerp(tempPoints[j], tempPoints[j + 1], t);
                    }
                }

                return tempPoints[0];
            }

            private static Vector3 CircularInterpolation(Vector3[] controlPoints, float t)
            {
                if (controlPoints.Length < 2) throw new System.ArgumentException("Need at least two points for circular interpolation");

                Vector3 center = (controlPoints[0] + controlPoints[1]) / 2;
                float radius = Vector3.Distance(controlPoints[0], controlPoints[1]) / 2;
                float angle = t * Mathf.PI * 2;

                return new Vector3(center.x + radius * Mathf.Cos(angle), center.y, center.z + radius * Mathf.Sin(angle));
            }
            public static Vector3 PointOnCircle(Plane plane, float radius, float t)
            {
                // Vector3[] points = new Vector3[numPoints + 1];
                Quaternion rotation = Quaternion.LookRotation(plane.X, plane.Normal); 


                float precentage = t;
                float angle = 2 * Mathf.PI * t;
                Vector3 point = new Vector3(Mathf.Cos(angle) * radius, 0f, Mathf.Sin(angle) * radius);
                Vector3 points = rotation * point + plane.basePoint;


                return points;
            }
        }
        public class bezier
        {
            public static Vector3[] generateBezier(Vector3[] controlPoints, int resolution, bool closed)
            {
                if (closed)
                {
                    Array.Resize(ref controlPoints, controlPoints.Length + 1);
                    controlPoints[controlPoints.Length - 1] = controlPoints[0];
                }
                Vector3[] samplePts = new Vector3[resolution + 1];
                for (int resCount = 0; resCount < resolution + 1; resCount++)
                {

                    Vector3[] pts = new Vector3[controlPoints.Length];
                    for (int i = 0; i < controlPoints.Length; i++)
                    {
                        pts[i] = controlPoints[i];
                    }

                    samplePts[resCount] = Curves.bezier.Bezier(pts, (float)resCount / (float)resolution);
                }
                return samplePts;

            }

            public static Vector3 Bezier(Vector3[] controlPoints, float t)
            {
                int n = controlPoints.Length - 1;
                t = Mathf.Clamp01(t);
                float oneMinusT = 1f - t;
                Vector3[] tempPoints = new Vector3[n + 1];
                for (int i = 0; i <= n; i++)
                {
                    tempPoints[i] = controlPoints[i];
                }
                for (int j = 1; j <= n; j++)
                {
                    for (int i = 0; i <= n - j; i++)
                    {
                        tempPoints[i] = oneMinusT * tempPoints[i] + t * tempPoints[i + 1];
                    }
                }
                return tempPoints[0];
            }
        }
        public static Vector3[] GeneratePointsOnCircle(Plane plane, float radius, int numPoints)
        {
            Vector3[] points = new Vector3[numPoints+1];
            Quaternion rotation = Quaternion.LookRotation(plane.X, plane.Normal);

            for (int i = 0; i < numPoints; i++)
            {
                float precentage = (i - numPoints);
                float angle = i * (2 * Mathf.PI / numPoints);
                Vector3 point = new Vector3(Mathf.Cos(angle) * radius, 0f, Mathf.Sin(angle) * radius);
                points[i] = rotation * point + plane.basePoint;
            }
            points[numPoints] = points[0];
            return points;
        }
        public static Vector3[] GenerateRectangleOnPlane(Plane plane, float x ,float y)
        {
            Quaternion rotation = Quaternion.LookRotation(plane.Normal, plane.X);
            Vector3[] points = new Vector3[5];
            points[0] = new Vector3(-x, -y);
            points[1] = new Vector3(-x, y);
            points[2] = new Vector3(x, y);
            points[3] = new Vector3(x, -y);
            points[4] = new Vector3(-x, -y);
            for (int i = 0; i < points.Length; i++)
            {
                points[i] = rotation * points[i] + plane.basePoint;
            }
            return points;
        }
        public static Vector3 GetPointOnCurve(Vector3[] controlPoints, int interpolationType, float t)
        {
            if (interpolationType == 0)
            {
                // polyline interpolation
                int index = Mathf.Min(Mathf.FloorToInt(t * controlPoints.Length), controlPoints.Length - 1);
                float localT = t * controlPoints.Length - index;

                return Vector3.Lerp(controlPoints[index], controlPoints[Mathf.Min(index + 1, controlPoints.Length - 1)], localT);
            }
            else if (interpolationType == 1)
            {
                // bezier interpolation
                return BezierInterpolation(controlPoints, t);
            }
            else
            {
                throw new System.ArgumentException("Invalid interpolation type");
            }
        }

        public static Vector3 GetTangentOnCurve(Vector3[] controlPoints, int interpolationType, float t)
        {
            Vector3 tangent;

            if (interpolationType == 0)
            {
                // polyline tangent
                int index = Mathf.Min(Mathf.FloorToInt(t * controlPoints.Length), controlPoints.Length - 1);
                tangent = controlPoints[Mathf.Min(index + 1, controlPoints.Length - 1)] - controlPoints[index];
            }
            else if (interpolationType == 1)
            {
                // bezier tangent
                Vector3 left = BezierInterpolation(controlPoints, Mathf.Max(0, t - 0.01f));
                Vector3 right = BezierInterpolation(controlPoints, Mathf.Min(1, t + 0.01f));

                tangent = (right - left).normalized;
            }

            
            else
            {
                throw new System.ArgumentException("Invalid interpolation type");
            }

            return tangent;
        }

        private static Vector3 BezierInterpolation(Vector3[] controlPoints, float t)
        {
            Vector3[] tempPoints = new Vector3[controlPoints.Length];
            System.Array.Copy(controlPoints, tempPoints, controlPoints.Length);

            for (int i = controlPoints.Length - 1; i > 0; i--)
            {
                for (int j = 0; j < i; j++)
                {
                    tempPoints[j] = Vector3.Lerp(tempPoints[j], tempPoints[j + 1], t);
                }
            }

            return tempPoints[0];
        }

        public static Vector3[] GeneratePolyline(int sides, float radius, Transform center)
        {
            float angle = 360.0f / sides;
            Vector3[] points = new Vector3[sides + 1];

            for (int i = 0; i < sides; i++)
            {
                float x = radius * Mathf.Sin(Mathf.Deg2Rad * (i * angle));
                float z = radius * Mathf.Cos(Mathf.Deg2Rad * (i * angle));
                Vector3 point = new Vector3(x, 0.0f, z);
                point = center.rotation * point;

                points[i] = center.position + point;

            }
            points[sides] = points[0];
            return points;
        }

    }
}
