using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;


namespace PlugPeaNodes
{
    public class PMesh
    {
        public static class CatmullClarkSubdivision
        {
           public static Mesh Subdivide(Mesh originalMesh, int subdivisions)
            {
                Mesh mesh = originalMesh;

                for (int i = 0; i < subdivisions; i++)
                {
                    mesh = SubdivideOnce(mesh);
                }

                return mesh;
            }

            public static Mesh SubdivideOnce(Mesh inputMesh)
            {
                Mesh subdividedMesh = new Mesh();

                Vector3[] vertices = inputMesh.vertices;
                Vector3[] normals = inputMesh.normals;
                int[] triangles = inputMesh.triangles;

                int numVertices = vertices.Length;
                int numTriangles = triangles.Length / 3;
                Vector3[] newVertices = new Vector3[numVertices + numTriangles];
                Vector3[] newNormals = new Vector3[numVertices + numTriangles];

                // Implement Catmull-Clark subdivision algorithm here
                // You need to calculate newVertices, newNormals, and update triangles

                // Set the new vertices, normals, and triangles for the subdivided mesh
                subdividedMesh.vertices = newVertices;
                subdividedMesh.normals = newNormals;
                subdividedMesh.triangles = triangles; // Replace with subdivided triangles.

                subdividedMesh.RecalculateBounds();
                subdividedMesh.RecalculateTangents();

                return subdividedMesh;
            }
        }
        public static Mesh GenerateSphere(Vector3 position, float radius, int latitudeSegments, int longitudeSegments)
        {
            float maxLongitude = Mathf.PI * 2;

            Mesh mesh = new Mesh();

            // Vertices
            Vector3[] vertices = new Vector3[(latitudeSegments + 1) * (longitudeSegments + 1)];
            int index = 0;

            for (int lat = 0; lat <= latitudeSegments; lat++)
            {
                float latitude = Mathf.PI / 2 - lat * Mathf.PI / latitudeSegments;
                float dy = Mathf.Sin(latitude) * radius;
                float dxz = Mathf.Cos(latitude) * radius;

                for (int lon = 0; lon <= longitudeSegments; lon++)
                {
                    float longitude = lon * maxLongitude / longitudeSegments;
                    float dx = Mathf.Cos(longitude) * dxz;
                    float dz = Mathf.Sin(longitude) * dxz;

                    Vector3 vertex = new Vector3(dx, dy, dz) + position;
                    vertices[index++] = vertex;
                }
            }

            // Triangles
            int[] triangles = new int[latitudeSegments * longitudeSegments * 6];
            index = 0;

            for (int lat = 0; lat < latitudeSegments; lat++)
            {
                for (int lon = 0; lon < longitudeSegments; lon++)
                {
                    int current = lat * (longitudeSegments + 1) + lon;
                    int next = current + longitudeSegments + 1;

                    triangles[index++] = current;
                    triangles[index++] = next + 1;
                    triangles[index++] = next;

                    triangles[index++] = current;
                    triangles[index++] = current + 1;
                    triangles[index++] = next + 1;
                }
            }

            // Assign vertices and triangles to the mesh
            mesh.vertices = vertices;
            mesh.triangles = triangles;

            // Recalculate the normals so lighting works correctly
            mesh.RecalculateNormals();

            return mesh;
        }
        public static Mesh ExtrudeCurve(Curve curve, Vector3 direction)
        {
            Vector3[] pts = curve.ControlPoints;
            
            Vector3[,] extrusionArray = new Vector3[ pts.Length, 2];
            for(int i = 0; i < 2; i++)
            {
                for(int j = 0; j < pts.Length; j++) 
                {
                    extrusionArray[j, i] = pts[j] + i * direction;

                }
            }
            
           Mesh m =  CreateMeshFromVertices(extrusionArray);



            return m;
        }
        public static Mesh CreateMeshFromVertices(Vector3[,] vertices)
        {
            int width = vertices.GetLength(0);
            int height = vertices.GetLength(1);

            Mesh mesh = new Mesh();
            List<Vector3> vertexList = new List<Vector3>();
            List<int> indexList = new List<int>();

            // Add vertices to vertexList and set up triangles using indexList
            for (int i = 0; i < width - 1; i++)
            {
                for (int j = 0; j < height - 1; j++)
                {
                    int baseIndex = vertexList.Count;

                    vertexList.Add(vertices[i, j]);
                    vertexList.Add(vertices[i + 1, j]);
                    vertexList.Add(vertices[i, j + 1]);
                    vertexList.Add(vertices[i + 1, j + 1]);

                    indexList.Add(baseIndex);
                    indexList.Add(baseIndex + 2);
                    indexList.Add(baseIndex + 1);

                    indexList.Add(baseIndex + 1);
                    indexList.Add(baseIndex + 2);
                    indexList.Add(baseIndex + 3);
                }
            }

            mesh.vertices = vertexList.ToArray();
            mesh.triangles = indexList.ToArray();
            mesh.RecalculateNormals();

            return mesh;
        }
        public static Mesh GenerateRectangularMesh(float x, float y, int resolutionX, int resolutionY, Plane plane)
        {
            Mesh mesh = new Mesh();
            Quaternion rotation = Quaternion.LookRotation(plane.X, plane.Normal  *-1);

            // Calculate half of the width and half of the length
            float halfWidth = x / 2f;
            float halfLength = y / 2f;

            // Calculate the step size between vertices
            float stepX = x / (resolutionX - 1);
            float stepY = y / (resolutionY - 1);

            // Create vertices
            Vector3[] vertices = new Vector3[resolutionX * resolutionY];
            for (int i = 0; i < resolutionX; i++)
            {
                for (int j = 0; j < resolutionY; j++)
                {
                    // Calculate the position relative to the center of the mesh
                    float offsetX = i * stepX - halfWidth;
                    float offsetZ = j * stepY - halfLength;

                    // Rotate and position the vertex on the plane
                    vertices[i * resolutionY + j] = rotation * new Vector3(offsetX, 0f, offsetZ) + plane.basePoint;
                }
            }

            // Create triangles (indexing the vertices)
            int[] triangles = new int[(resolutionX - 1) * (resolutionY - 1) * 6];
            int t = 0;
            for (int i = 0; i < resolutionX - 1; i++)
            {
                for (int j = 0; j < resolutionY - 1; j++)
                {
                    int topLeft = i * resolutionY + j;
                    int topRight = (i + 1) * resolutionY + j;
                    int bottomLeft = i * resolutionY + j + 1;
                    int bottomRight = (i + 1) * resolutionY + j + 1;

                    triangles[t++] = topLeft;
                    triangles[t++] = bottomLeft;
                    triangles[t++] = topRight;

                    triangles[t++] = topRight;
                    triangles[t++] = bottomLeft;
                    triangles[t++] = bottomRight;
                }
            }

            // Assign vertices and triangles to the mesh
            mesh.vertices = vertices;
            mesh.triangles = triangles;

            // Calculate normals (you might want to adjust this based on your usage)
            mesh.RecalculateNormals();

            return mesh;
        }
        public static Mesh GenerateCube(float x, float y, float z, Plane plane)
        {
            Mesh mesh = new Mesh();

            // Calculate half of the width, length, and height
            float halfWidth = x / 2f;
            float halfLength = y / 2f;
            float halfHeight = z / 2f;

            // Calculate the local space rotation based on the plane
            Quaternion rotation = Quaternion.LookRotation(plane.Normal, Vector3.up);

            // Create vertices
            Vector3[] vertices = new Vector3[8];
            vertices[0] = rotation * new Vector3(-halfWidth, -halfLength, -halfHeight) + plane.ClosestPointOnPlane(plane.basePoint - plane.Normal * halfHeight);
            vertices[1] = rotation * new Vector3(halfWidth, -halfLength, -halfHeight) + plane.ClosestPointOnPlane(plane.basePoint - plane.Normal * halfHeight);
            vertices[2] = rotation * new Vector3(-halfWidth, -halfLength, halfHeight) + plane.ClosestPointOnPlane(plane.basePoint - plane.Normal * halfHeight);
            vertices[3] = rotation * new Vector3(halfWidth, -halfLength, halfHeight) + plane.ClosestPointOnPlane(plane.basePoint - plane.Normal * halfHeight);
            vertices[4] = rotation * new Vector3(-halfWidth, halfLength, -halfHeight) + plane.ClosestPointOnPlane(plane.basePoint - plane.Normal * halfHeight);
            vertices[5] = rotation * new Vector3(halfWidth, halfLength, -halfHeight) + plane.ClosestPointOnPlane(plane.basePoint - plane.Normal * halfHeight);
            vertices[6] = rotation * new Vector3(-halfWidth, halfLength, halfHeight) + plane.ClosestPointOnPlane(plane.basePoint - plane.Normal * halfHeight);
            vertices[7] = rotation * new Vector3(halfWidth, halfLength, halfHeight) + plane.ClosestPointOnPlane(plane.basePoint - plane.Normal * halfHeight);

            // Create triangles
            int[] triangles = new int[]
            {
        0, 1, 2,  // Front
        1, 3, 2,
        4, 6, 5,  // Back
        5, 6, 7,
        0, 2, 4,  // Bottom
        4, 2, 6,
        1, 5, 3,  // Top
        5, 7, 3,
        0, 4, 1,  // Left
        1, 4, 5,
        2, 3, 6,  // Right
        3, 7, 6
            };

            // Assign vertices and triangles to the mesh
            mesh.vertices = vertices;
            mesh.triangles = triangles;

            // Calculate normals (you might want to adjust this based on your usage)
            mesh.RecalculateNormals();

            return mesh;
        }



        public static Mesh JoinMeshes(params Mesh[] meshes)
        {
            // Create a new mesh to hold the combined meshes
            Mesh combinedMesh = new Mesh();

            // Create lists to hold the combined vertices and triangles
            List<Vector3> combinedVertices = new List<Vector3>();
            List<int> combinedTriangles = new List<int>();

            foreach (Mesh mesh in meshes)
            {
                // Ensure the mesh isn't null
                if (mesh != null)
                {
                    int vertexOffset = combinedVertices.Count;

                    // Add the vertices from the mesh to the combined list
                    combinedVertices.AddRange(mesh.vertices);

                    // Add the triangles from the mesh to the combined list, offset by the current vertex count
                    foreach (int triangleVertexIndex in mesh.triangles)
                    {
                        combinedTriangles.Add(triangleVertexIndex + vertexOffset);
                    }
                }
            }

            // Assign the combined vertices and triangles to the mesh
            combinedMesh.SetVertices(combinedVertices);
            combinedMesh.SetTriangles(combinedTriangles, 0);

            // Recalculate the bounds and normals for the new mesh
            combinedMesh.RecalculateBounds();
            combinedMesh.RecalculateNormals();

            // Return the combined mesh
            return combinedMesh;
        }
        public static Mesh FlipMesh(Mesh originalMesh, bool flipInsideOut)
        {
            // Create a new mesh to hold the flipped mesh
            Mesh flippedMesh = new Mesh();

            // Duplicate the original mesh
            flippedMesh.vertices = originalMesh.vertices;
            flippedMesh.triangles = originalMesh.triangles;

            // Reverse the order of the triangle indices to flip the mesh inside out
            if (flipInsideOut)
            {
                for (int i = 0; i < flippedMesh.subMeshCount; i++)
                {
                    int[] triangles = flippedMesh.GetTriangles(i);
                    for (int j = 0; j < triangles.Length; j += 3)
                    {
                        int temp = triangles[j];
                        triangles[j] = triangles[j + 2];
                        triangles[j + 2] = temp;
                    }
                    flippedMesh.SetTriangles(triangles, i);
                }
            }

            // Flip the face normals of the mesh
            flippedMesh.RecalculateNormals();
            Vector3[] normals = flippedMesh.normals;
            for (int i = 0; i < normals.Length; i++)
            {
                normals[i] = -normals[i];
            }
            flippedMesh.normals = normals;

            // Flip the tangents of the mesh
            Vector4[] tangents = flippedMesh.tangents;
            for (int i = 0; i < tangents.Length; i++)
            {
                tangents[i].w = -tangents[i].w;
            }
            flippedMesh.tangents = tangents;

            // Return the flipped mesh
            return flippedMesh;

        }
        public static Mesh MeshLoft(List<Vector3[]> curves, bool reverse)
        {
            if (curves != null)
            {
                int count;
                try
                {
                     count = curves[0].Length;
                }
                catch { count = 0; }
                if (!curves.All(crv => crv.Length == count))
                {
                    Debug.LogError("All input arrays must have the same length.");
                    return null;
                }

                if (reverse)
                {
                    for (int i = 0; i < curves.Count; i++)
                    {
                        System.Array.Reverse(curves[i]);
                    }
                }

                var mesh = new Mesh();
                var vertices = new Vector3[count * curves.Count];
                var triangles = new List<int>();

                for (int i = 0; i < curves.Count; i++)
                {
                    for (int j = 0; j < count; j++)
                    {
                        vertices[i * count + j] = curves[i][j];
                    }
                }

                for (int i = 0; i < curves.Count - 1; i++)
                {
                    for (int j = 0; j < count - 1; j++)
                    {
                        triangles.Add(i * count + j);
                        triangles.Add((i + 1) * count + j);
                        triangles.Add((i + 1) * count + j + 1);

                        triangles.Add(i * count + j);
                        triangles.Add((i + 1) * count + j + 1);
                        triangles.Add(i * count + j + 1);
                    }
                }

                mesh.vertices = vertices;

                mesh.triangles = triangles.ToArray();
                mesh.RecalculateNormals();

                return mesh;
            }
            else { return null;}
        }
        public static Mesh DoubleSide(Mesh originalMesh)
        {
            Mesh newmesh = new Mesh();
            newmesh.vertices = originalMesh.vertices;
            
            newmesh.triangles = originalMesh.triangles.Reverse().ToArray();
            return (newmesh);
        }
            //public static Mesh DoubleSide(Mesh originalMesh)
            //{
            //    // Create a new mesh to hold the double-sided mesh
            //    Mesh doubleSidedMesh = new Mesh();

            //    // Duplicate the vertices of the original mesh
            //    Vector3[] vertices = originalMesh.vertices;
            //    int numVertices = vertices.Length;
            //    Vector3[] newVertices = new Vector3[numVertices * 2];
            //    for (int i = 0; i < numVertices; i++)
            //    {
            //        newVertices[i] = vertices[i];
            //        newVertices[i + numVertices] = vertices[i];
            //    }

            //    // Duplicate the normals of the original mesh
            //    Vector3[] normals = originalMesh.normals;
            //    Vector3[] newNormals = new Vector3[numVertices * 2];
            //    for (int i = 0; i < numVertices; i++)
            //    {
            //        newNormals[i] = normals[i];
            //        newNormals[i + numVertices] = -normals[i];
            //    }

            //    // Create a new set of triangles that represent the mesh as if it were double-sided
            //    int[] triangles = originalMesh.triangles;
            //    int numTriangles = triangles.Length;
            //    int[] newTriangles = new int[numTriangles * 2];
            //    for (int i = 0; i < numTriangles; i += 3)
            //    {
            //        // First set of triangles
            //        newTriangles[i] = triangles[i];
            //        newTriangles[i + 1] = triangles[i + 1];
            //        newTriangles[i + 2] = triangles[i + 2] + numVertices;

            //        // Second set of triangles
            //        newTriangles[i + numTriangles] = triangles[i] + numVertices;
            //        newTriangles[i + numTriangles + 1] = triangles[i + 2] + numVertices;
            //        newTriangles[i + numTriangles + 2] = triangles[i + 1] + numVertices;
            //    }

            //    // Calculate the normals for the interior faces
            //    for (int i = 0; i < numTriangles; i += 3)
            //    {
            //        // First set of triangles
            //        int normalIndex = triangles[i];
            //        Vector3 normal = newNormals[normalIndex];
            //        if (Vector3.Dot(normal, newVertices[triangles[i]] - newVertices[triangles[i + 1]]) > 0)
            //        {
            //            normal = -normal;
            //        }
            //        newNormals[normalIndex] = normal;

            //        normalIndex = triangles[i + 1];
            //        normal = newNormals[normalIndex];
            //        if (Vector3.Dot(normal, newVertices[triangles[i + 1]] - newVertices[triangles[i + 2]]) > 0)
            //        {
            //            normal = -normal;
            //        }
            //        newNormals[normalIndex] = normal;

            //        normalIndex = triangles[i + 2] + numVertices;
            //        normal = newNormals[normalIndex];
            //        if (Vector3.Dot(normal, newVertices[triangles[i + 2] + numVertices] - newVertices[triangles[i + 1] + numVertices]) > 0)
            //        {
            //            normal = -normal;
            //        }
            //        newNormals[normalIndex] = normal;

            //        // Second set of triangles



            //        normalIndex = triangles[i + 2] + numVertices;
            //        normal = newNormals[normalIndex];
            //        if (Vector3.Dot(normal, newVertices[triangles[i + 2] + numVertices] - newVertices[triangles[i] + numVertices]) > 0)
            //        {
            //            normal = -normal;
            //        }
            //        newNormals[normalIndex] = normal;

            //        normalIndex = triangles[i + 1] + numVertices;
            //        normal = newNormals[normalIndex];
            //        if (Vector3.Dot(normal, newVertices[triangles[i + 1] + numVertices] - newVertices[triangles[i + 2] + numVertices]) > 0)
            //        {
            //            normal = -normal;
            //        }
            //        newNormals[normalIndex] = normal;
            //    }

            //    // Assign the new vertices, normals, and triangles to the double-sided mesh
            //    doubleSidedMesh.vertices = newVertices;
            //    doubleSidedMesh.normals = newNormals;
            //   // doubleSidedMesh.RecalculateNormals();
            //    doubleSidedMesh.triangles = newTriangles;

            //    return doubleSidedMesh;
            //}



        }



    public static class MeshUtils
    {
        public static Mesh WeldMesh(Mesh originalMesh, float threshold)
        {
            Mesh mesh = new Mesh();

            Vector3[] oldVerts = originalMesh.vertices;
            Vector3[] newVerts;
            List<int> newTris = new List<int>();
            List<Vector3> verts = new List<Vector3>();

            for (int i = 0; i < originalMesh.triangles.Length; i++)
            {
                Vector3 vert = oldVerts[originalMesh.triangles[i]];

                bool found = false;
                for (int j = 0; j < verts.Count; j++)
                {
                    if (Vector3.Distance(vert, verts[j]) < threshold)
                    {
                        newTris.Add(j);
                        found = true;
                        break;
                    }
                }

                if (!found)
                {
                    verts.Add(vert);
                    newTris.Add(verts.Count - 1);
                }
            }

            newVerts = verts.ToArray();

            mesh.vertices = newVerts;
            mesh.triangles = newTris.ToArray();
            mesh.RecalculateNormals();
            mesh.RecalculateBounds();

            return mesh;
        }
    }




}

